import { Component, Input, OnDestroy, OnInit, ViewChild } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { Store } from '@ngxs/store';
import * as pipelineActions from 'app/store/pipelines.action';
import { PipelinesState } from 'app/store/pipelines.state';
import { cloneDeep } from 'lodash';
import { SemanticModalComponent } from 'ng-semantic/ng-semantic';
import { DragulaService } from 'ng2-dragula';
import { Subscription } from 'rxjs';
import { finalize, first, flatMap } from 'rxjs/operators';
import { Job } from '../../../../model/job.model';
import { AllKeys } from '../../../../model/keys.model';
import { PermissionValue } from '../../../../model/permission.model';
import { Pipeline } from '../../../../model/pipeline.model';
import { Project } from '../../../../model/project.model';
import { Stage } from '../../../../model/stage.model';
import { KeyService } from '../../../../service/keys/keys.service';
import { PipelineCoreService } from '../../../../service/pipeline/pipeline.core.service';
import { VariableService } from '../../../../service/variable/variable.service';
import { ActionEvent } from '../../../../shared/action/action.event.model';
import { AutoUnsubscribe } from '../../../../shared/decorator/autoUnsubscribe';
import { ToastService } from '../../../../shared/toast/ToastService';

@Component({
    selector: 'app-pipeline-workflow',
    templateUrl: './pipeline.workflow.html',
    styleUrls: ['./pipeline.workflow.scss']
})
@AutoUnsubscribe()
export class PipelineWorkflowComponent implements OnInit, OnDestroy {

    @Input() project: Project;
    @Input('currentPipeline')
    set currentPipeline(data: Pipeline) {
        this.pipeline = cloneDeep(data);
        this.originalPipeline = this.pipeline;
        if (!this.pipeline) {
            return;
        }

        if (!this.selectedJob || this.pipeline.forceRefresh) {
            this.selectDefaultJob();
        }

        if (!this.pipeline.stages) {
            this.pipeline.stages = new Array<Stage>();
        }

        if (this.pipeline.preview) {
            this.previewMode = true;
            this.pipeline = this.pipeline.preview;
            this.selectDefaultJob();
            this._pipCoreService.toggleAsCodeEditor({ open: false, save: false });
        } else {
            let jobFound = false;
            let stageFound = null;
            if (this.selectedStage) {
                stageFound = this.selectedStage && data.stages && data.stages.find((stage) => stage.id === this.selectedStage.id);
                jobFound = stageFound && this.selectedJob && stageFound.jobs && stageFound.jobs.some((job) => {
                    return job.pipeline_action_id === this.selectedJob.pipeline_action_id;
                });
                if (stageFound && !jobFound) {
                    this.selectDefaultJobInStage(this.selectedStage);
                }
            }
            if (!stageFound && !jobFound) {
                this.selectDefaultJob();
            }
            this.previewMode = false;
        }
    }
    @Input() queryParams: {};

    @ViewChild('editStageModal')
    editStageModal: SemanticModalComponent;

    pipeline: Pipeline;
    selectedStage: Stage;
    selectedJob: Job;
    suggest: Array<string>;
    permissionValue = PermissionValue;
    originalPipeline: Pipeline;
    keys: AllKeys;
    previewMode = false;

    loadingStage = false;

    pipelinePreviewSubscription: Subscription;
    asCodeEditorSubscription: Subscription;
    dragulaSubscription: Subscription;

    constructor(
        private store: Store,
        private _toast: ToastService,
        private _dragularService: DragulaService,
        private _translate: TranslateService,
        private _varService: VariableService,
        private _pipCoreService: PipelineCoreService,
        private _keyService: KeyService
    ) {
        this._dragularService.createGroup('bag-stage', {
            moves: function (el, source, handle) {
                return handle.classList.contains('move');
            },
            accepts: function (el, target, source, sibling) {
                if (sibling === null) {
                    return false;
                }
                return true;
            }
        });
        this.dragulaSubscription = this._dragularService.drop('bag-stage').subscribe(({ el, source }) => {
            setTimeout(() => {
                let stageMovedBuildOrder = Number(el.id.replace('step', ''));
                let stageMoved: Stage;
                for (let i = 0; i < this.pipeline.stages.length; i++) {
                    if (this.pipeline.stages[i].build_order === stageMovedBuildOrder) {
                        stageMoved = this.pipeline.stages[i];
                        stageMoved.build_order = i + 1;
                        break;
                    }
                }
                this.store.dispatch(new pipelineActions.MovePipelineStage({
                    projectKey: this.project.key,
                    pipelineName: this.pipeline.name,
                    stage: stageMoved
                })).subscribe(() => this._toast.success('', this._translate.instant('pipeline_stage_moved')));
            });
        });
    }

    selectDefaultJob() {
        if (this.pipeline.stages && this.pipeline.stages.length &&
            this.pipeline.stages[0].jobs && this.pipeline.stages[0].jobs.length) {
            this.selectJob(this.pipeline.stages[0].jobs[0], this.pipeline.stages[0]);
        }
    }

    selectDefaultJobInStage(stage: Stage) {
        if (stage.jobs && stage.jobs.length) {
            this.selectJob(stage.jobs[0], stage);
        }
    }

    ngOnDestroy() {
        this._dragularService.destroy('bag-stage');
    }

    /**
     * Init selected stage + pipeline date
     */
    ngOnInit() {
        if (this.queryParams && this.queryParams['stage']) {
            this.selectedStage = cloneDeep(this.pipeline.stages.find(s => s.name === this.queryParams['stage']));
        }
        if (this.pipeline.stages && this.pipeline.stages.length > 0 && !this.selectedStage) {
            this.selectedStage = cloneDeep(this.pipeline.stages[0]);
        }

        this._keyService.getAllKeys(this.project.key).subscribe(k => {
            this.keys = k;
        })

        this._varService.getContextVariable(this.project.key, this.pipeline.id).pipe(first()).subscribe(s => this.suggest = s);

    }

    addStageAndJob(): void {
        let s = new Stage();
        s.enabled = true;
        if (!this.pipeline.stages) {
            this.pipeline.stages = new Array<Stage>();
        }

        s.name = 'Stage ' + (this.pipeline.stages.length + 1);

        this.store.dispatch(new pipelineActions.AddPipelineStage({
            projectKey: this.project.key,
            pipelineName: this.pipeline.name,
            stage: s
        })).pipe(
            flatMap(() => this.store.selectOnce(PipelinesState.selectPipeline(this.project.key, this.pipeline.name)))
        ).subscribe((pip) => {
            this._toast.success('', this._translate.instant('stage_added'));
            this.selectedStage = pip.stages[pip.stages.length - 1];
            this.addJob(this.selectedStage);
        });
    }

    addJob(s: Stage): void {
        let jobToAdd = new Job();
        jobToAdd.action.name = 'New Job';
        jobToAdd.enabled = true;
        jobToAdd.pipeline_stage_id = s.id;
        this.store.dispatch(new pipelineActions.AddPipelineJob({
            projectKey: this.project.key,
            pipelineName: this.pipeline.name,
            stageId: s.id,
            job: jobToAdd
        })).pipe(
            flatMap(() => this.store.selectOnce(PipelinesState.selectPipeline(this.project.key, this.pipeline.name)))
        )
            .subscribe((pip) => {
                this._toast.success('', this._translate.instant('stage_job_added'));
                let currentStage = pip.stages.find((stage) => this.selectedStage.id === stage.id);
                if (currentStage) {
                    this.selectJob(currentStage.jobs[currentStage.jobs.length - 1], this.selectedStage);
                }
            });
    }

    /**
     * Event on stage
     * @param type Type of event (update/delete)
     */
    stageEvent(type: string): void {
        this.loadingStage = true;
        switch (type) {
            case 'update':
                this.store.dispatch(new pipelineActions.UpdatePipelineStage({
                    projectKey: this.project.key,
                    pipelineName: this.pipeline.name,
                    changes: this.selectedStage
                })).pipe(finalize(() => this.loadingStage = false))
                    .subscribe(() => {
                        this._toast.success('', this._translate.instant('stage_updated'));
                        this.editStageModal.hide();
                    });
                break;
            case 'delete':
                this.store.dispatch(new pipelineActions.DeletePipelineStage({
                    projectKey: this.project.key,
                    pipelineName: this.pipeline.name,
                    stage: this.selectedStage
                })).pipe(finalize(() => this.loadingStage = false))
                    .subscribe(() => {
                        this._toast.success('', this._translate.instant('stage_deleted'));
                        this.editStageModal.hide();
                        this.selectedStage = null;
                        this.selectedJob = null;
                    });
                break;
        }
    }

    openEditModal(s: Stage): void {
        this.selectedStage = cloneDeep(s);
        if (this.editStageModal) {
            this.editStageModal.show({ autofocus: false, closable: false, observeChanges: true });
        }
    }

    selectStage(s: Stage): void {
        this.selectedStage = cloneDeep(s);
    }

    selectJob(j: Job, s: Stage): void {
        this.selectStage(s);
        this.selectedJob = cloneDeep(j);
    }

    /**
     * Manage action from jobs
     * @param event
     */
    jobEvent(event: ActionEvent): void {
        let job: Job = cloneDeep(this.selectedJob);
        job.action = event.action;
        job.action.loading = true;
        job.enabled = event.action.enabled;
        if (job.action.actions) {
            job.action.actions.forEach(a => {
                if (a.parameters) {
                    a.parameters.forEach(p => {
                        p.value = p.value.toString();
                    });
                }
            });
        }
        if (job.action.parameters) {
            job.action.parameters.forEach(p => {
                p.value = p.value.toString();
            });
        }

        switch (event.type) {
            case 'update':
                this.store.dispatch(new pipelineActions.UpdatePipelineJob({
                    projectKey: this.project.key,
                    pipelineName: this.pipeline.name,
                    stageId: this.selectedStage.id,
                    changes: job
                })).pipe(finalize(() => {
                    job.action.loading = false;
                    job.action.hasChanged = false;
                })).subscribe(() => {
                    this._toast.success('', this._translate.instant('stage_job_updated'));
                    this.selectedStage = this.pipeline.stages.find((s) => this.selectedStage.id === s.id) || this.selectedStage;
                    this.selectedJob = this.selectedStage.jobs.find((j) => this.selectedJob.action.id === j.action.id) || this.selectedJob;
                });
                break;
            case 'delete':
                this.store.dispatch(new pipelineActions.DeletePipelineJob({
                    projectKey: this.project.key,
                    pipelineName: this.pipeline.name,
                    stageId: this.selectedStage.id,
                    job
                })).pipe(finalize(() => job.action.loading = false))
                    .subscribe(() => {
                        this._toast.success('', this._translate.instant('stage_job_deleted'));
                        this.selectedJob = undefined;
                    });
                break;
        }
    }

    savePreview() {
        this.previewMode = false;
        this._pipCoreService.toggleAsCodeEditor({ open: false, save: true });
    }

    showAsCodeEditor() {
        this._pipCoreService.toggleAsCodeEditor({ open: true, save: false });
    }
}
